import java.io.*;
// jtux is the library that gives the POSIX interface.
import jtux.*;

// This is an interface to the SystemV semaphores that are
// easier to use. The SystemV interface to semaphores is
// very weird and hard to understand. If you're interested
// you can take a look here, but you do not need to understand
// the implementation, just the interface.
abstract class Semaphore {

    // Create a new semaphore. value should be its initial
    // value. It should be 1 in most cases. It returns a
    // handle to the semaphore.
    static int create(int value) {
	int semid = 0;
	try {
	    semid = USysVIPC.semget(UConstant.IPC_PRIVATE,
				    1,
				    UConstant.IPC_CREAT | 00400 | 00200);
	    USysVIPC.u_semun_int arg = new USysVIPC.u_semun_int();
	    arg.val = value;
	    USysVIPC.semctl(semid, 0, UConstant.SETVAL, arg);
	} catch (Exception e) {
	    System.out.print("Exception: ");
	    System.out.println(">" + e.toString() + "<");
	    System.exit(1);
	}
	return semid;
    }

    // Destroy a semaphore. This will free the semaphore given
    // as argument. You should always do this when you're finished
    // with a semaphore.
    static void destroy(int sem) {
	try {
	    USysVIPC.u_semun semun = new USysVIPC.u_semun();
	    USysVIPC.semctl(sem, 0, UConstant.IPC_RMID, semun);
	} catch (Exception e) {
	    System.out.print("Exception: ");
	    System.out.println(">" + e.toString() + "<");
	    System.exit(1);
	}
    }

    // Wait for the semaphore. It will block until the semaphore
    // is available.
    static void wait(int sem) {
	USysVIPC.s_sembuf wait[] = new USysVIPC.s_sembuf[1];
	wait[0] = new USysVIPC.s_sembuf();
	wait[0].sem_num = 0;
	wait[0].sem_op = -1;
	wait[0].sem_flg = 010000;
	try {
	    USysVIPC.semop(sem, wait, 1);
	} catch (Exception e) {
	    	    System.out.print("Exception: ");
	    System.out.println(">" + e.toString() + "<");
	    System.exit(1);
	}
    }

    // Signal on the semaphore. When we leave a critical section
    // we need to use this to make the semaphore available to
    // others.
    static void signal(int sem) {
	USysVIPC.s_sembuf signal[] = new USysVIPC.s_sembuf[1];
	signal[0] = new USysVIPC.s_sembuf();
	signal[0].sem_num = 0;
	signal[0].sem_op = 1;
	signal[0].sem_flg = 010000;
	try {
	    USysVIPC.semop(sem, signal, 1);
	} catch (Exception e) {
	    	    System.out.print("Exception: ");
	    System.out.println(">" + e.toString() + "<");
	    System.exit(1);
	}
    }

    // The same as wait, but will not block if the semaphore
    // is not available. Instead it will return 0, as opposed
    // to 1 which is what it normally returns.
    // Depending on how you solve the deadlock freedom problem,
    // you may or may not need to use this.
    static int noblockwait(int sem) {
	USysVIPC.s_sembuf wait[] = new USysVIPC.s_sembuf[1];
	wait[0] = new USysVIPC.s_sembuf();
	wait[0].sem_num = 0;
	wait[0].sem_op = -1;
	wait[0].sem_flg = UConstant.IPC_NOWAIT;
	try {
	    USysVIPC.semop(sem, wait, 1);
	} catch (UErrorException e) {
	    if (e.getCode() == UConstant.EAGAIN)
		return 0;
	    else {
		System.out.print("Exception: ");
		System.out.println(">" + e.toString() + "<");
		System.exit(1);
	    }
	}
	return 1;
    }
}